"""
Generates a Swagger definition for registered endpoints.

Note that:
 -  Swagger operations and type names use different conventions from the internal definitions
    because we want to make usage friendly for code generation (e.g. bravado)

 -  Marshmallow to JSON Schema conversion is somewhat simplistic. There are several projects
    that already implement this conversion that we could try adapting. At the moment, the
    overhead of adapting another library's conventions is too high.

 -  All resource definitions are assumed to be shared and are declared in the "definitions"
    section of the result.

 -  All errors are defined generically.


"""
from werkzeug.routing import BuildError

from openapi import model as swagger

from microcosm_flask.conventions.registry import (
    get_qs_schema,
    get_request_schema,
    get_response_schema,
)
from microcosm_flask.errors import ErrorSchema, ErrorContextSchema, SubErrorSchema
from microcosm_flask.naming import name_for
from microcosm_flask.routing import make_path
from microcosm_flask.swagger.naming import operation_name, type_name
from microcosm_flask.swagger.schema import build_parameter, build_schema


def build_swagger(graph, ns, operations):
    """
    Build out the top-level swagger definition.

    """
    base_path = make_path(graph, ns.path)
    schema = swagger.Swagger(
        swagger="2.0",
        info=swagger.Info(
            title=graph.metadata.name,
            version=ns.version,
        ),
        consumes=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        produces=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        basePath=base_path,
        paths=swagger.Paths(),
        definitions=swagger.Definitions(),
    )
    add_paths(schema.paths, base_path, operations)
    add_definitions(schema.definitions, operations)
    schema.validate()
    return schema


def add_paths(paths, base_path, operations):
    """
    Add paths to swagger.

    """
    for operation, ns, rule, func in operations:
        path = build_path(operation, ns)
        if not path.startswith(base_path):
            continue
        method = operation.value.method.lower()
        paths.setdefault(
            path[len(base_path):],
            swagger.PathItem(),
        )[method] = build_operation(operation, ns, rule, func)


def add_definitions(definitions, operations):
    """
    Add definitions to swagger.

    """
    # general error schema per errors.py
    for error_schema_class in [ErrorSchema, ErrorContextSchema, SubErrorSchema]:
        error_schema = error_schema_class()
        definitions[type_name(name_for(error_schema))] = swagger.Schema(build_schema(error_schema))

    # add all request and response schemas
    for operation, obj, rule, func in operations:
        request_schema = get_request_schema(func)
        if request_schema:
            definitions.setdefault(type_name(name_for(request_schema)), swagger.Schema(build_schema(request_schema)))

        response_schema = get_response_schema(func)
        if response_schema:
            definitions.setdefault(type_name(name_for(response_schema)), swagger.Schema(build_schema(response_schema)))


def build_path(operation, ns):
    """
    Build a path URI for an operation.

    """
    try:
        return ns.url_for(operation)
    except BuildError as error:
        uri_templates = {
            argument: "{{{}}}".format(argument)
            for argument in error.suggested.arguments
        }
        return ns.url_for(operation, **uri_templates)


def body_param(schema):
    return swagger.BodyParameter(**{
        "name": "body",
        "in": "body",
        "schema": swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(schema))),
        }),
    })


def query_param(name, field, required=False):
    """
    Build a query parameter definition.

    """
    parameter = build_parameter(field)
    parameter["name"] = name
    parameter["in"] = "query"
    parameter["required"] = False

    return swagger.QueryParameterSubSchema(**parameter)


def path_param(name, param_type="string"):
    """
    Build a path parameter definition.

    """
    return swagger.PathParameterSubSchema(**{
        "name": name,
        "in": "path",
        "required": True,
        "type": param_type,
    })


def build_operation(operation, ns, rule, func):
    """
    Build an operation definition.

    """
    swagger_operation = swagger.Operation(
        operationId=operation_name(operation, ns),
        parameters=swagger.ParametersList([
        ]),
        responses=swagger.Responses(),
        tags=[ns.subject_name],
    )

    # path parameters
    swagger_operation.parameters.extend([
        # TODO: inject type information for parameters based on converter syntax
        path_param(argument)
        for argument in rule.arguments
    ])

    # query string parameters
    qs_schema = get_qs_schema(func)
    if qs_schema:
        swagger_operation.parameters.extend([
            query_param(name, field)
            for name, field in qs_schema.fields.items()
        ])

    # body parameter
    request_schema = get_request_schema(func)
    if request_schema:
        swagger_operation.parameters.append(
            body_param(request_schema)
        )

    add_responses(swagger_operation, operation, ns, func)
    return swagger_operation


def add_responses(swagger_operation, operation, ns, func):
    """
    Add responses to an operation.

    """
    # default error
    swagger_operation.responses["default"] = build_response(
        description="An error occcurred",
        resource=type_name(name_for(ErrorSchema())),
    )

    if getattr(func, "__doc__", None):
        description = func.__doc__.strip().splitlines()[0]
    else:
        description = "{} {}".format(operation.value.name, ns.subject_name)

    # resources response
    swagger_operation.responses[str(operation.value.default_code)] = build_response(
        description=description,
        resource=get_response_schema(func),
    )


def build_response(description, resource=None):
    """
    Build a response definition.

    """
    response = swagger.Response(
        description=description,
    )
    if resource is not None:
        response.schema = swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(resource))),
        })
    return response



"""
Generate JSON Schema for Marshmallow schemas.

"""
from marshmallow import fields

from microcosm_flask.fields import EnumField
from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name


# see: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py
FIELD_MAPPINGS = {
    fields.Dict: ("object", None),
    fields.Integer: ("integer", "int32"),
    fields.Number: ("number", None),
    fields.Float: ("number", "float"),
    fields.Decimal: ("number", None),
    fields.String: ("string", None),
    fields.Boolean: ("boolean", None),
    fields.UUID: ("string", "uuid"),
    fields.DateTime: ("string", "date-time"),
    fields.Date: ("string", "date"),
    fields.Time: ("string", None),
    fields.Email: ("string", "email"),
    fields.URL: ("string", "url"),
    fields.Field: ("object", None),
    fields.Raw: ("object", None),
    fields.List: ("array", None),
    fields.Nested: (None, None),
    fields.Method: ("object", None),
    EnumField: ("string", None),
}


def build_parameter(field):
    """
    Build a parameter from a marshmallow field.

    See: https://github.com/marshmallow-code/apispec/blob/dev/apispec/ext/marshmallow/swagger.py#L81

    """
    field_type, field_format = FIELD_MAPPINGS[type(field)]
    parameter = {}
    if field_type:
        parameter["type"] = field_type
    if field.metadata.get("description"):
        parameter["description"] = field.metadata["description"]
    if field_format:
        parameter["format"] = field_format
    if field.default:
        parameter["default"] = field.default

    # enums
    enum = getattr(field, "enum", None)
    if enum:
        parameter["enum"] = [choice.name for choice in enum]

    # nested
    if isinstance(field, fields.Nested):
        parameter["$ref"] = "#/definitions/{}".format(type_name(name_for(field.schema)))

    # arrays
    if isinstance(field, fields.List):
        parameter["items"] = build_parameter(field.container)

    return parameter


def build_schema(marshmallow_schema):
    """
    Build JSON schema from a marshmallow schema.

    """
    fields = [
        (name, marshmallow_schema.fields[name])
        for name in sorted(marshmallow_schema.fields.keys())
    ]
    schema = {
        "type": "object",
        "properties": {
            field.dump_to or name: build_parameter(field)
            for name, field in fields
        },
        "required": [
            field.dump_to or name
            for name, field in fields
            if field.required
        ]
    }
    return schema



"""
Audit log support for Flask routes.

"""
from logging import getLogger
from functools import wraps
from json import loads
from traceback import format_exc

from flask import current_app, g, request

from microcosm_flask.errors import (
    extract_context,
    extract_error_message,
    extract_status_code,
)


def audit(func):
    """
    Record a Flask route function in the audit log.

    Generates a JSON record in the Flask log for every request.

    """
    logger = getLogger("audit")

    @wraps(func)
    def wrapper(*args, **kwargs):
        response = None

        # always include these fields
        audit_dict = dict(
            operation=request.endpoint,
            func=func.__name__,
            method=request.method,
        )

        # include request body on debug (if any)
        if current_app.debug and request.get_json(force=True, silent=True):
            audit_dict["request_body"] = request.get_json(force=True)

        try:
            response = func(*args, **kwargs)
        except Exception as error:
            status_code = extract_status_code(error)
            success = 0 < status_code < 400
            audit_dict.update(
                success=success,
                message=extract_error_message(error)[:2048],
                context=extract_context(error),
                stack_trace=None if success else format_exc(limit=10),
                status_code=status_code,
            )
            raise
        else:
            body, status_code = parse_response(response)

            audit_dict.update(
                success=True,
                status_code=status_code,
            )

            # include response body on debug (if any)
            if current_app.debug and body and not g.get("hide_body"):
                try:
                    audit_dict["response_body"] = loads(body)
                except (TypeError, ValueError):
                    # not json
                    audit_dict["response_body"] = body

            return response
        finally:
            # always log at INFO; we can't know whether a raised exception
            # is an error or expected behavior
            logger.info(audit_dict)

    return wrapper


def parse_response(response):
    """
    Parse a Flask response into a body and status code.

    The returned value from a Flask view could be:
        * a tuple of (response, status) or (response, status, headers)
        * a Response object
        * a string
    """
    if isinstance(response, tuple) and len(response) > 1:
        return response[0], response[1]
    try:
        return response.data, response.status_code
    except AttributeError:
        return response, 200


def configure_audit_decorator(graph):
    """
    Configure the audit decorator.

    Example Usage:

        @graph.audit
        def login(username, password):
            ...
    """
    return audit



"""
A naming convention and discovery mechanism for HTTP endpoints.

Operations provide a naming convention for references between endpoints,
allowing easy construction of links or audit trails for external consumption.

"""
from collections import namedtuple

from enum import Enum, unique


# metadata for an operation
OperationInfo = namedtuple("OperationInfo", ["name", "method", "pattern", "default_code"])


NODE_PATTERN = "{}.{}"
EDGE_PATTERN = "{}.{}.{}"


@unique
class Operation(Enum):
    """
    An enumerated set of operation types, which know how to resolve themselves into
    URLs and hrefs.

    """
    # discovery operation
    Discover = OperationInfo("discover", "GET", NODE_PATTERN, 200)

    # collection operations
    Search = OperationInfo("search", "GET", NODE_PATTERN, 200)
    Create = OperationInfo("create", "POST", NODE_PATTERN, 201)
    # bulk update is possible here with PATCH

    # instance operations
    Retrieve = OperationInfo("retrieve", "GET", NODE_PATTERN, 200)
    Delete = OperationInfo("delete", "DELETE", NODE_PATTERN, 204)
    Replace = OperationInfo("replace", "PUT", NODE_PATTERN, 200)
    Update = OperationInfo("update", "PATCH", NODE_PATTERN, 200)

    # relation operations
    CreateFor = OperationInfo("create_for", "POST", EDGE_PATTERN, 201)
    RetrieveFor = OperationInfo("retrieve_for", "GET", EDGE_PATTERN, 200)
    SearchFor = OperationInfo("search_for", "GET", EDGE_PATTERN, 200)

    # ad hoc operations
    Command = OperationInfo("command", "POST", NODE_PATTERN, 200)
    Query = OperationInfo("query", "GET", NODE_PATTERN, 200)

    @classmethod
    def from_name(cls, name):
        for operation in cls:
            if operation.value.name.lower() == name.lower():
                return operation
        else:
            raise ValueError(name)
